// FileCache.cpp: implementation of the FileCache class.
//
//////////////////////////////////////////////////////////////////////

#include "FileCache.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

FileCache::FileCache (int iMaxNumBytes) {

	m_iMaxNumBytes = iMaxNumBytes;
	m_iNumBuckets = (int) ((double) m_iMaxNumBytes / (AVERAGE_FILE_LENGTH));

	if (m_iNumBuckets == 0) {
		m_iNumBuckets = 1;
	}

	m_iNumFiles = 0;
	m_iNumBytes = 0;
	m_ppBucket = new StringLinkedList<File>* [m_iNumBuckets];

	m_ppBucketMutex = new Mutex* [m_iNumBuckets];
	m_ppCounterMutex = new Mutex* [m_iNumBuckets];
	m_piWriters = new int [m_iNumBuckets];
	m_piReaders = new int [m_iNumBuckets];
	
	m_pStatMutex = new Mutex();
	m_pExitMutex = new Mutex();

	m_pReadList = new StringLinkedList<File>();
	m_pReadMutex = new Mutex();

	m_pFifoList = new StringLinkedList<File>();
	m_pFifoMutex = new Mutex();

	for (int i = 0; i < m_iNumBuckets; i ++) {
		m_ppBucket[i] = NULL;
		m_ppBucketMutex[i] = new Mutex();
		m_ppCounterMutex[i] = new Mutex();
		m_piWriters[i] = 0;
		m_piReaders[i] = 0;
	}

	// Start new threads
	m_bExit = false;
	
	m_pSizeControlThread = new Thread();
	m_pRefreshControlThread = new Thread();

	m_pSizeControlThread->Start (SizeControlThread, this, THREAD_PRIORITY_LOWEST);
	m_pRefreshControlThread->Start (RefreshControlThread, this, THREAD_PRIORITY_LOWEST);
}

FileCache::~FileCache() {

	// Notify threads of exit
	m_pExitMutex->WaitForSingleObject();
	m_bExit = true;
	m_pExitMutex->ReleaseMutex();

	// Wait for threads to exit
	m_pSizeControlThread->WaitForSingleObject();
	delete m_pSizeControlThread;

	m_pRefreshControlThread->WaitForSingleObject();
	delete m_pRefreshControlThread;

	// Delete data
	for (int i = 0; i < m_iNumBuckets; i ++) {
		if (m_ppBucket != NULL && m_ppBucket[i] != NULL) {
			delete m_ppBucket[i];
		}

		if (m_ppBucketMutex != NULL && m_ppBucketMutex[i] != NULL) {
			delete m_ppBucketMutex[i];
		}

		if (m_ppCounterMutex != NULL && m_ppCounterMutex[i] != NULL) {
			delete m_ppCounterMutex[i];
		}
	}

	if (m_ppBucket != NULL) {
		delete [] m_ppBucket;
	}

	if (m_ppBucketMutex != NULL) {
		delete [] m_ppBucketMutex;
	}

	if (m_ppCounterMutex != NULL) {
		delete [] m_ppCounterMutex;
	}

	if (m_piWriters != NULL) {
		delete [] m_piWriters;
	}

	if (m_pReadList != NULL) {
		char* pszName;
		File* pszData;
		while (m_pReadList->GetFirstElement (&pszName, &pszData) != NULL) {
			m_pReadList->DeleteFirstElement();
		}
		delete m_pReadList;
	}

	if (m_pReadMutex != NULL) {
		delete m_pReadMutex;
	}

	if (m_pFifoList != NULL) {
		char* pszName;
		File* pszData;
		while (m_pFifoList->GetFirstElement (&pszName, &pszData) != NULL) {
			m_pFifoList->DeleteFirstElement();
		}
		delete m_pFifoList;
	}

	if (m_pFifoMutex != NULL) {
		delete m_pFifoMutex;
	}

	if (m_piReaders != NULL) {
		delete [] m_piReaders;
	}

	if (m_pStatMutex != NULL) {
		delete m_pStatMutex;
	}

	if (m_pExitMutex != NULL) {
		delete m_pExitMutex;
	}
}


bool FileCache::Exit() {
	m_pExitMutex->WaitForSingleObject();
	bool bExit = m_bExit;
	m_pExitMutex->ReleaseMutex();
	return bExit;
}


File* FileCache::GetFile (char* pszFileName) {

	if (pszFileName == NULL || pszFileName [strlen (pszFileName) - 1] == '\\') {
		return NULL;
	}

	File* pfFile = this->RetrieveFile (pszFileName);

	if (pfFile == NULL) {
	
		// Open file and add it to list
		pfFile = new File();
		if (pfFile->Open (pszFileName) == OK) {
			this->AddFile (pszFileName, pfFile);
		} else {
			delete pfFile;
			return NULL;
		}
	} else {

		// Add to read list
		m_pReadMutex->WaitForSingleObject();
		m_pReadList->AddElement (pszFileName, pfFile);
		m_pReadMutex->ReleaseMutex();
	}

	// Increment reference count
	pfFile->AddRef();

	return pfFile;
}

File* FileCache::GetFileFromCache (char* pszFileName) {

	if (pszFileName == NULL || pszFileName [strlen (pszFileName) - 1] == '\\') {
		return NULL;
	}

	File* pfFile = this->RetrieveFile (pszFileName);

	if (pfFile != NULL) {
		pfFile->AddRef();
	}

	return pfFile;
}

void FileCache::ReleaseFile (File* pfFile) {
	if (pfFile != NULL) {
		pfFile->Release();
	}
}


File* FileCache::RetrieveFile (char* pszFileName) {
	
	// Readers algorithm
	int iHash = this->GetHashValue (pszFileName);

	// Increment readers and get number of writers
	m_ppCounterMutex[iHash]->WaitForSingleObject();
	m_piReaders[iHash] ++;
	int iWriters = m_piWriters[iHash];
	m_ppCounterMutex[iHash]->ReleaseMutex();
	
	// Get bucket lock if there are writers
	if (iWriters > 0) {
		m_ppCounterMutex[iHash]->WaitForSingleObject();
		m_piReaders[iHash] --;
		m_ppCounterMutex[iHash]->ReleaseMutex();
		m_ppBucketMutex[iHash]->WaitForSingleObject();
		m_ppCounterMutex[iHash]->WaitForSingleObject();
		m_piReaders[iHash] ++;
		m_ppCounterMutex[iHash]->ReleaseMutex();
	}

	// Critical section:  read File object from list
	File* pfRetVal = (m_ppBucket[iHash] == NULL) ? NULL : m_ppBucket[iHash]->GetElement (pszFileName);

	// Unlock if we locked before
	if (iWriters > 0) {
		m_ppBucketMutex[iHash]->ReleaseMutex();
	}

	// Decrement readers
	m_ppCounterMutex[iHash]->WaitForSingleObject();
	m_piReaders[iHash] --;
	m_ppCounterMutex[iHash]->ReleaseMutex();

	return pfRetVal;
}


void FileCache::AddFile (char* pszFileName, File* pfFile) {

	// Writers algorithm
	int iReaders = 1, iHash = this->GetHashValue (pszFileName);

	// Increment writers
	m_ppCounterMutex[iHash]->WaitForSingleObject();
	m_piWriters[iHash] ++;
	m_ppCounterMutex[iHash]->ReleaseMutex();

	// Get bucket lock
	m_ppBucketMutex[iHash]->WaitForSingleObject();

	// Spinlock until all the readers are gone
	while (true) {
		
		// Get number of readers
		m_ppCounterMutex[iHash]->WaitForSingleObject();
		iReaders = m_piReaders[iHash];
		m_ppCounterMutex[iHash]->ReleaseMutex();

		// Spinlock if still > 0
		if (iReaders > 0) {
			Mutex::SpinLock();
		} else {
			break;
		}
	}

	// Critical section:  manipulate linked list
	if (m_ppBucket[iHash] == NULL) {
		m_ppBucket[iHash] = new StringLinkedList<File>();
	}
	m_ppBucket[iHash]->AddElement (pszFileName, pfFile);

	// Unlock bucket
	m_ppBucketMutex[iHash]->ReleaseMutex();

	// Decrement writers
	m_ppCounterMutex[iHash]->WaitForSingleObject();
	m_piWriters[iHash] --;
	m_ppCounterMutex[iHash]->ReleaseMutex();

	// Update statistics
	m_pStatMutex->WaitForSingleObject();
	m_iNumFiles ++;
	m_iNumBytes += pfFile->GetNumBytes();
	m_pStatMutex->ReleaseMutex();

	// Add file to tail of FIFO list of files
	m_pFifoMutex->WaitForSingleObject();
	m_pFifoList->AddElement (pszFileName, pfFile);
	m_pFifoMutex->ReleaseMutex();
}


int FileCache::GetHashValue (char* pszFileName) {
	
	int i, iHash = 0, iLength = strlen (pszFileName);
	
	char* pszNewName = new char [iLength + 1];
	sprintf (pszNewName, pszFileName);
	char* pszLowerName = _strlwr (pszNewName);
	
	for (i = 0; i < iLength; i ++) {
		iHash += (int) pszLowerName[i];
	}
	
	delete [] pszNewName;
	
	return iHash % m_iNumBuckets;
}


int FileCache::SizeControlThread (void* lpvData) {

	FileCache* pThis = (FileCache*) lpvData;
	Mutex* pStatMutex = pThis->m_pStatMutex;
	Mutex* pFifoMutex = pThis->m_pFifoMutex;
	StringLinkedList<File>* pFifoList = pThis->m_pFifoList;

	int iNumBytes, iMaxNumBytes, iHash, iReaders, iFileSize;
	char* pszFileName;
	char* pszFileNameCopy;
	File* pfFile = NULL;

	// Loop forever
	while (true) {
		
		// Get current size and current max size of cache
		pStatMutex->WaitForSingleObject();
		iNumBytes = pThis->m_iNumBytes;
		iMaxNumBytes = pThis->m_iMaxNumBytes;
		pStatMutex->ReleaseMutex();

		// Adjust cache size if necessary
		while (iNumBytes > iMaxNumBytes) {

			// Lock FIFO list
			pFifoMutex->WaitForSingleObject();
			
			// Get file information
			pFifoList->GetFirstElement (&pszFileNameCopy, &pfFile);

			if (pszFileName == NULL || pfFile == NULL) {

				// Unlock FIFO list
				pThis->m_pFifoMutex->ReleaseMutex();
				break;
			}

			pszFileName = new char [strlen (pszFileNameCopy) + 1];
			sprintf (pszFileName, pszFileNameCopy);

			iFileSize = pfFile->GetNumBytes();

			// Delete file from beginning of FIFO list
			pFifoList->DeleteFirstElement();

			// Unlock FIFO list
			pFifoMutex->ReleaseMutex();
			
			// Writers algorithm
			iReaders = 1;
			iHash = pThis->GetHashValue (pszFileName);
			
			// Increment writers
			pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
			pThis->m_piWriters[iHash] ++;
			pThis->m_ppCounterMutex[iHash]->ReleaseMutex();
			
			// Get bucket lock
			pThis->m_ppBucketMutex[iHash]->WaitForSingleObject();
			
			// Spinlock until all the readers are gone
			while (true) {
				
				// Get number of readers
				pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
				iReaders = pThis->m_piReaders[iHash];
				pThis->m_ppCounterMutex[iHash]->ReleaseMutex();
				
				// Spinlock if still > 0
				if (iReaders > 0) {
					Mutex::SpinLock();
				} else {
					break;
				}
			}

			// Spin until file has zero references
			while (pfFile->GetNumRefs() > 0) {
				Mutex::SpinLock();
			}

			// Delete file from bucket
			pThis->m_ppBucket[iHash]->DeleteElement (pszFileName);

			// Unlock bucket
			pThis->m_ppBucketMutex[iHash]->ReleaseMutex();

			// Decrement writers
			pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
			pThis->m_piWriters[iHash] --;
			pThis->m_ppCounterMutex[iHash]->ReleaseMutex();

			// Update statistics
			pStatMutex->WaitForSingleObject();
			pThis->m_iNumBytes -= iFileSize;
			iNumBytes = pThis->m_iNumBytes;
			iMaxNumBytes = pThis->m_iMaxNumBytes;
			pStatMutex->ReleaseMutex();
		
			delete [] pszFileName;
		}

		// Check for exit
		if (pThis->Exit()) {
			break;
		}

		// Rest
		Utility::Sleep (ONE_SECOND);
	}

	return OK;
}


int FileCache::RefreshControlThread (void* lpvData) {

	FileCache* pThis = (FileCache*) lpvData;
	StringLinkedList<File>* pReadList = pThis->m_pReadList;
	Mutex* pReadMutex = pThis->m_pReadMutex;
	Mutex* pStatMutex = pThis->m_pStatMutex;

	File* pfFile;
	File* pfNewFile;
	char* pszFileName;
	char pszName [MAX_FILENAME_LENGTH];

	struct _stat stFile;
	int iCurSize, iNewSize, iCurTime, iNewTime, iReaders, iHash;

	// Loop forever
	while (true) {

		// Lock read list
		pReadMutex->WaitForSingleObject();
		
		if (pReadList->GetFirstElement (&pszFileName, &pfFile)) {
			
			// Copy the name
			sprintf (pszName, pszFileName);

			// Erase the first element and release the mutex
			pReadList->DeleteFirstElement();
			pReadMutex->ReleaseMutex();
			
			// Get the file pointer
			pfFile = pThis->GetFileFromCache (pszName);
			
			if (pfFile != NULL) {

				// Release if file is too big for its data to sit in the filecache
				iCurSize = pfFile->GetNumBytes();
				if (iCurSize == 0) {
					pfFile->Release();
				} else {
					
					// Get the last modified times
					_stat (pszName, &stFile);
					iCurTime = pfFile->GetLastModifiedIntTime();
					iNewTime = stFile.st_mtime;
					
					// Release if no change
					if (iCurTime == iNewTime) {
						pfFile->Release();
					} else {
						
						// Replace the file
						iReaders = 1;
						iHash = pThis->GetHashValue (pszName);
						
						// Increment writers
						pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
						pThis->m_piWriters[iHash] ++;
						pThis->m_ppCounterMutex[iHash]->ReleaseMutex();
						
						// Get bucket lock
						pThis->m_ppBucketMutex[iHash]->WaitForSingleObject();
						
						// Spinlock until all the readers are gone
						while (true) {
							
							// Get number of readers
							pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
							iReaders = pThis->m_piReaders[iHash];
							pThis->m_ppCounterMutex[iHash]->ReleaseMutex();
							
							// Spinlock if still > 0
							if (iReaders > 0) {
								Mutex::SpinLock();
							} else {
								break;
							}
						}
						
						// Spin until file has only one reference (ours)
						while (pfFile->GetNumRefs() > 1) {
							Mutex::SpinLock();
						}
						
						// Replace file with new file
						pThis->m_ppBucket[iHash]->DeleteElement (pszName);
						pfNewFile = new File();
						if (pfNewFile->Open (pszName) == OK) {
							pThis->m_ppBucket[iHash]->AddElement (pszName, pfNewFile);
							iNewSize = pfNewFile->GetNumBytes();
						} else {
							delete pfNewFile;
							iNewSize = 0;
						}
						
						// Unlock bucket
						pThis->m_ppBucketMutex[iHash]->ReleaseMutex();
						
						// Decrement writers
						pThis->m_ppCounterMutex[iHash]->WaitForSingleObject();
						pThis->m_piWriters[iHash] --;
						pThis->m_ppCounterMutex[iHash]->ReleaseMutex();
						
						// Update statistics
						pStatMutex->WaitForSingleObject();
						pThis->m_iNumBytes += (iNewSize - iCurSize);
						pStatMutex->ReleaseMutex();
					}
				}
			}

		} else {
			pReadMutex->ReleaseMutex();
			Utility::Sleep (ONE_SECOND);
		}

		// Check for exit
		if (pThis->Exit()) {
			break;
		}
	}

	return OK;
}